<?php
require_once ALI_APPLICATION_PATH . '/services/AbstractService.php';

class Ali_Service_Linker extends Ali_Service_AbstractService
{
    const PREFIX = Ali_WpPlugin::PREFIX;
    const TAG_OPEN  = '<ali_temp>';
    const TAG_CLOSE  = '</ali_temp>';

    protected $_keywordSlugCache;
    protected $_urlCache;

    //replaceCallback properties
    protected $_replaceLinkAttrs;
    protected $_replaceBaseUrl;
    protected $_replaceCount;
    protected $_replaceLimit;

    public function parseKeywordsField($text)
    {
        $urlCache = array();
        $keywordSlugCache = array();

        $lines = explode("\n", $text);

        foreach ($lines as $lineIndex => $line) {
            $line = trim($line);
            if ('' == $line) {
                continue;
            }

            $items = explode(',', $line);
            $keywords = array();
            $urls = array();

            foreach ($items as $item) {
                $item = trim($item);
                if ('' == $item) {
                    continue;
                }

                if (!parse_url($item, PHP_URL_SCHEME)) {
                  $keywords[] = mb_strtolower($item);
                } else {
                  $urls[] = $item;
                }
            }

            if (!$urls) {
                require_once ALI_APPLICATION_PATH . '/services/Linker/Exception.php';
                throw new Ali_Service_Linker_Exception('No URL found in line #' . ($lineIndex+1));
            }
            if (!$keywords) {
                require_once ALI_APPLICATION_PATH . '/services/Linker/Exception.php';
                throw new Ali_Service_Linker_Exception('No keywords found in line #' . ($lineIndex+1));
            }

            foreach ($keywords as $keyword) {
                if (!array_key_exists($keyword, $keywordSlugCache)) {
                    $keywordSlug = trim(mb_strtolower($keyword));
                    $keywordSlug = preg_replace('~[^\w\d]+~u', '-', $keywordSlug);
                    $keywordSlug = trim($keywordSlug, '-');
                    $keywordSlug = urlencode($keywordSlug);
                    if (in_array($keywordSlug, $keywordSlugCache)) {
                        $keywordSlug .= '--' . (count($keywordSlugCache) + 1);
                    }
                    $keywordSlugCache[$keyword] = $keywordSlug;
                } else {
                    $keywordSlug = $keywordSlugCache[$keyword];
                }

                if (!isset($urlCache[$keywordSlug])) {
                    $urlCache[$keywordSlug] = array_values(array_unique($urls));
                } else {
                    $urlCache[$keywordSlug] = array_values(array_unique(array_merge($urlCache[$keywordSlug], $urls)));
                }
            }
        }

        $this->_keywordSlugCache = $keywordSlugCache;
        $this->_urlCache = $urlCache;

        return $this;
    }

    public function loadKeywordSlugCache()
    {
        if (is_null($this->_keywordSlugCache)) {
            $options = Ali_Application::getModel('Options');

            if (!($this->_keywordSlugCache = Ali_Application::getModel('Options')->getValue('keyword_slug_cache'))) {
                $this->_keywordSlugCache = array();
            }
        }

        return $this;
    }

    public function loadUrlCache()
    {
        if (is_null($this->_urlCache)) {
            if (!($this->_urlCache = Ali_Application::getModel('Options')->getValue('url_cache'))) {
                $this->_urlCache = array();
            }
        }

        return $this;
    }

    public function saveCache()
    {
        $options = Ali_Application::getModel('Options');

        $options->setValue('keyword_slug_cache', $this->_keywordSlugCache);
        $options->setValue('url_cache', $this->_urlCache);

        return $this;
    }

    public function filterContent($content)
    {
        $this->loadKeywordSlugCache();
        if (!$this->_keywordSlugCache) {
            return $content;
        }

        $keywords = array();
        foreach (array_keys($this->_keywordSlugCache) as $keyword) {
            $keywords[] = preg_quote($keyword, '~');
            $escapedKeyword = htmlspecialchars($keyword, ENT_NOQUOTES, 'UTF-8');
            if ($escapedKeyword != $keyword) {
                $keywords[] = preg_quote($escapedKeyword, '~');
            }
        }
        $keywords = implode('|', $keywords);

        $content = preg_replace("~(^|[^\w\d])({$keywords})($|[^\w\d])~isu",
            '${1}' . self::TAG_OPEN . '${2}' . self::TAG_CLOSE . '${3}',
            $content
        );

        // clean temp tag in <>
        $content = $this->_multiPregReplace(
            '~(<[^>]*?)'
            . self::TAG_OPEN . '(.*?)' . self::TAG_CLOSE
            . '([^>]*?>)~isu',
            '${1}${2}${3}',
            $content
        );

        // clean temp tag in restricted tags
        foreach (Ali_Application::getConfig()->linker->restrictedTags as $tag) {
            $content = $this->_removeKeywordMarkFromTag($tag, $content);
        }

        // Prepare callback:
        $options = Ali_Application::getModel('Options');

        $this->_replaceLinkAttrs = '';
        if ($options->getValue('link_new_window')) {
            $this->_replaceLinkAttrs .= ' target="_blank"';
        }
        if ($options->getValue('link_nofollow')) {
            $this->_replaceLinkAttrs .= ' rel="nofollow"';
        }

        $path = array(site_url());

        global $wp_rewrite;
        if (!$wp_rewrite->using_mod_rewrite_permalinks()) {
            $path[] = $wp_rewrite->index;
        }

        $path[] = $options->getValue('dir_path');

        $this->_replaceBaseUrl = implode('/', $path) . '/';
        $this->_replaceCount = 0;
        if (is_null($this->_replaceLimit)) {
            $this->_initReplaceLimit();
        }

        // Exec replace
        $content = preg_replace_callback('~' . self::TAG_OPEN . '(.*?)' . self::TAG_CLOSE . '~isu',
            array($this, 'replaceCallback'), $content,
            $this->_replaceLimit? $this->_replaceLimit : -1);

        // clean other temp tags
        $content = preg_replace('~' . self::TAG_OPEN . '(.*?)' . self::TAG_CLOSE . '~isu',
           '${1}',
           $content
        );

        $this->_replaceLimit = NULL;

        return $content;
    }

    public function setPost($post)
    {
        $this->_initReplaceLimit($post->ID);
    }

    protected function _initReplaceLimit($postId=NULL)
    {
        if (is_null($postId) && is_singular()) {
            $postId = get_the_ID();
        }

        if (!is_null($postId)) {
            $meta = Ali_Application::getModel('PostMeta')->setPostId(get_the_ID());
            if ($meta->getValue('custom')) {
                $this->_replaceLimit = (int) $meta->getValue('max_links');
                return;
            }
        }

        $this->_replaceLimit = (int) Ali_Application::getModel('Options')->getValue('max_links');
    }

    public function replaceCallback($matches)
    {
        $result = $matches[1];
        $keyword = mb_strtolower($result);

        if ($this->_replaceCount < $this->_replaceLimit
            || (!$this->_replaceLimit)
        ) {
            if (isset($this->_keywordSlugCache[$keyword])) {
                $slug = $this->_keywordSlugCache[$keyword];
            } else {
                $keyword = html_entity_decode($keyword, ENT_QUOTES, 'UTF-8');
                if (isset($this->_keywordSlugCache[$keyword])) {
                    $slug = $this->_keywordSlugCache[$keyword];
                } else {
                    return $result . "<!-- Cannot find a slug for keyword {$result} -->";
                }
            }

            $result = "<a{$this->_replaceLinkAttrs} href=\"{$this->_replaceBaseUrl}{$slug}\">{$result}</a>";

            $this->_replaceCount++;
        }

        return $result;
    }


    /**
    * Continues regex replace until pattern not found in subject
    *
    * @param string $pattern
    * @param string $subject
    */
    protected function _multiPregReplace($pattern, $replacement, $subject)
    {
        do {
            $subject = preg_replace($pattern, $replacement, $subject, -1, $count);
        } while ($count > 0);

        return $subject;
    }

    protected function _removeKeywordMarkFromTag($tag, $subject)
    {
        if (preg_match_all("~(<\s*{$tag}(?:\b[^>]*)?>)(.+?)(<\s*/\s*{$tag}\s*>)~isu",
            $subject, $matches, PREG_SET_ORDER)
        ) {
            foreach ($matches as $match) {
                $replace = preg_replace('~' . self::TAG_OPEN . '(.*?)' . self::TAG_CLOSE . '~isu',
                    '${1}', $match[2]);
                if ($replace != $match[2]) {
                    $subject = str_replace($match[0], $match[1] . $replace . $match[3], $subject);
                }
            }
        }

        return $subject;
    }

    public function redirect($slug)
    {
        if (headers_sent($file, $line)) {
            exit("Error: Headers already sent. Output started at {$file}:{$line}");
        }

        $this->loadUrlCache();
        $this->loadKeywordSlugCache();

        //$stat = Ali_Application::getModel('Stat');

        $keyword = array_search($slug, $this->_keywordSlugCache);
        if (FALSE === $keyword) {
            // Unknown keyword
            return $this;
        }

        //$lastUrl = $stat->lastUrl($keyword);

        if (!isset($this->_urlCache[$slug])) {
            // This doesn't have to happen
            return $this;
        }

        $urls = $this->_urlCache[$slug];

        $index = 0;
        if (count($urls) > 1) {
            $lastUrl = Ali_Application::getModel('Tracking')
                ->findLastUrlForKeyword($keyword);
            if (!is_null($lastUrl)) {
                $index = array_search($lastUrl, $urls);
                $index = (FALSE !== $index && $index < (count($urls)-1))
                    ? $index + 1
                    : 0;
            }
        }

        if (!isset($urls[$index])) {
            exit("Error: invalid URL index: {$index}");
        }

        $url = $urls[$index];

        $cfg = Ali_Application::getConfig()->tracking->cookie;
        $cookieName = $cfg->name;
        if (isset($_COOKIE[$cookieName])) {
            $visitor = $_COOKIE[$cookieName];
        } else {
            $visitor = uniqid();
            setcookie($cookieName, $visitor, time() + $cfg->lifetime, '/');
        }

        Ali_Application::getModel('Tracking')->recordClick(array(
            'keyword' => $keyword,
            'url' => $url,
            'visitor' => $visitor,
        ));

        header('Location: ' . $url, TRUE, 302);
        $encodedUrl = htmlspecialchars($url, ENT_QUOTES, 'utf-8');
echo
"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">
<html>
<head>
    <script type=\"text/javascript\">
        window.location=\"{$url}\";
    </script>
    <title>Redirect</title>
    <meta http-equiv=\"REFRESH\" content=\"0;url={$encodedUrl}\">
    <meta name=\"ROBOTS\" content=\"NOINDEX, NOFOLLOW\">
</head>
<body>
    Redirected to {$encodedUrl}.<br />
    <a href=\"{$encodedUrl}\">Click here</a> if your browser doesn't support redirection.
</body>
</html>";
        exit();
    }

    //
    // Getters & setters
    //

    public function getKeywordSlugCache()
    {
        return $this->_keywordSlugCache;
    }

    public function getUrlCache()
    {
        return $this->_urlCache;
    }
}